﻿#region Copyright
/* Copyright 2009-2010 Oliver Sturm <oliver@sturmnet.org> All rights reserved. */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Diagnostics;
using System.Threading.Tasks;

namespace TreeWalker {
  class Program {
    static void Main(string[] args) {
      var tree = CreateTree(42);

      // Just to check -------------------------------------------------
      WalkTree(tree, i => Console.Write("{0} ", i));
      Console.WriteLine( );

      // Now let's do something expensive and measure ------------------

      //var tree = CreateTree(10000);
      //Stopwatch stopwatch = new Stopwatch( );
      //stopwatch.Start( );

      //WalkTree(tree, i => Thread.SpinWait(300000));

      //stopwatch.Stop( );
      //Console.WriteLine("Elapsed: {0}", stopwatch.ElapsedMilliseconds);

      // What about some results? --------------------------------------

      //var tree = CreateTree(10000);
      //Stopwatch stopwatch = new Stopwatch( );
      //stopwatch.Start( );

      //var result = WalkTreeSumTasks(tree);
      //stopwatch.Stop( );

      //Console.WriteLine("Result: {0}", result);
      //Console.WriteLine("Elapsed: {0}", stopwatch.ElapsedMilliseconds);

    }

    private static void WalkTree(Tree<int> tree, Action<int> action) {
      if (tree.IsEmpty)
        return;

      WalkTree(tree.Left, action);
      WalkTree(tree.Right, action);

      action(tree.Value);
    }

    #region WalkTreeThreads
    private static void WalkTreeThreads(Tree<int> tree, Action<int> action) {
      if (tree.IsEmpty)
        return;

      var t1 = new Thread(o => WalkTreeThreads(tree.Left, action));
      t1.Start( );
      var t2 = new Thread(o => WalkTreeThreads(tree.Right, action));
      t2.Start( );

      t1.Join( );
      t2.Join( );

      action(tree.Value);
    }
    #endregion

    #region WalkTreeTasks
    private static void WalkTreeTasks(Tree<int> tree, Action<int> action) {
      if (tree.IsEmpty)
        return;

      var t1 = Task.Factory.StartNew(() => WalkTreeTasks(tree.Left, action));
      var t2 = Task.Factory.StartNew(() => WalkTreeTasks(tree.Right, action));

      Task.WaitAll(t1, t2);

      action(tree.Value);
    }
    #endregion

    #region WalkTreeSum
    private static int WalkTreeSum(Tree<int> tree) {
      if (tree.IsEmpty)
        return 0;

      Thread.SpinWait(100000);

      var left = WalkTreeSum(tree.Left);
      var right = WalkTreeSum(tree.Right);

      return tree.Value + left + right;
    }
    #endregion

    #region WalkTreeSumTasks
    private static int WalkTreeSumTasks(Tree<int> tree) {
      if (tree.IsEmpty)
        return 0;

      Thread.SpinWait(100000);

      var left = Task.Factory.StartNew<int>(() => WalkTreeSum(tree.Left));
      var right = Task.Factory.StartNew<int>(() => WalkTreeSum(tree.Right));

      return tree.Value + left.Result + right.Result;
    }
    #endregion


    #region CreateTree
    private static Tree<int> CreateTree(int count) {
      return CreateTree(1, count);
    }

    static Tree<int> CreateTree(int from, int to) {
      if (to <= from)
        return new Tree<int>(from);
      int thisVal = (to - from) / 2 + from;
      return new Tree<int>(thisVal,
        thisVal > from ? CreateTree(from, thisVal - 1) : Tree<int>.Empty,
        thisVal < to ? CreateTree(thisVal + 1, to) : Tree<int>.Empty);
    }
    #endregion
  }
}